/*
 * $Id: s_file.c,v 1.43 2006/03/26 05:50:50 slava Exp $
 *
 * Copyright (C) 2000-2006 by Slava Monich
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met: 
 *
 *   1.Redistributions of source code must retain the above copyright 
 *     notice, this list of conditions and the following disclaimer. 
 *   2.Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer 
 *     in the documentation and/or other materials provided with the 
 *     distribution. 
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) ARISING 
 * IN ANY WAY OUT OF THE USE OR INABILITY TO USE THIS SOFTWARE, EVEN 
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. 
 *
 * The views and conclusions contained in the software and documentation 
 * are those of the authors and should not be interpreted as representing 
 * any official policies, either expressed or implied.
 */

#include "s_fio.h"

/**
 * Initializes the shared portion of all file structures. The path parameter
 * may be NULL, which leaves name buffer empty, which in turn will cause
 * FILE_Name to query the target file name.
 */
Bool FILE_Init(File * f, Str path, Bool attach, IODesc io)
{
    f->io = io;
    f->flags = FILE_IS_OPEN | FILE_CAN_READ | FILE_CAN_WRITE;
    if (attach) f->flags |= FILE_IS_ATTACHED;

    STRBUF_Init(&f->name);
    STRBUF_Init(&f->buf);

    /* copy the file name. that's the only operation that can fail */
    return (path ? STRBUF_Copy(&f->name, path) : True);
}

/**
 * Open a file. Returns NULL if file cannot be opened or memory 
 * allocation fails.
 */
File * FILE_Open(Str path, const char * mode, IODesc io)
{
    File * f = NULL;

    if (!io) {
#ifdef  __KERNEL__
        ASSMSG("No default I/O in kernel mode!");
        return NULL;
#else /* ! __KERNEL__ */
        io = PlainFile;
#endif /* ! __KERNEL__ */
    }

    ASSERT(io->open);
    if (io->open) {
        f = (*io->open)(path, mode);
        if (f) {
            if (FILE_Init(f, path, False, io)) {
                return f;
            }
            f->io->close(f);
            f->io->free(f);
        }
    }

    return NULL;
}

/**
 * Reopens the file with specified mode. FILE_Close() is still necessary
 * even if this call fails.
 */
Bool FILE_Reopen(File * f, Str path, const char * mode)
{
    ASSERT(f);
    if (f && (f->flags & FILE_IS_OPEN) && 
        f->io->reopen && f->io->reopen(f, path, mode)) {
        if (STRBUF_Copy(&f->name, path)) {
            return True;
        } else {
            FILE_Finish(f); /* no more I/O */
        }
    }
    return False;
}

/**
 * Sets a file type specific I/O parameter.
 */
Bool FILE_SetParam(File * f, Str name, void * value)
{
    ASSERT(f);
    if (f && (f->flags & FILE_IS_OPEN) && f->io->setparam) {
        return f->io->setparam(f, name, value);
    }
    return False;
}

/**
 * Returns the file name. The returned pointer is valid until next 
 * FILE_Close(), FILE_Reopen() or FILE_Detach call. If this file
 * does not have a name (passed NULL path to FILE_Init) then the
 * name of the target file will be returned.
 */
Str FILE_Name(File * f)
{
    ASSERT(f);
    while (f) {
        if (STRBUF_Length(&f->name) > 0) {
            return STRBUF_GetString(&f->name);
        }
        f = FILE_Target(f);
    }

    /* return NULL? */
    return TEXT("(null)");
}

/**
 * Returns the flags
 */
int FILE_Flags(File * f)
{
    ASSERT(f);
    return (f->flags & FILE_PUBLIC_FLAGS);
}

/**
 * Set flags
 */
void FILE_SetFlag(File * f, int flags)
{
    ASSERT(f);
    f->flags |= (flags & FILE_PUBLIC_FLAGS);
}

/**
 * Clear flags
 */
void FILE_ClrFlag(File * f, int flags)
{
    ASSERT(f);
    f->flags &= (~(flags & FILE_PUBLIC_FLAGS));
}

/**
 * Reads data from the stream. Returns the number of bytes actually read,
 * which may be less than the requested count if an error occurs or if the
 * end of the file is encountered before reaching the requested count.
 * Returns (-1) if an error occurs and no data has been read.
 */
int FILE_Read(File * f, void * buf, int len)
{
    ASSERT(f);
    ASSERT(len >= 0);
    if (CAN_READ(f) && len >= 0) {
        int n1 = 0;
        I8u * dest = (I8u*)buf;
        while (f->pushed > 0 && n1 < len) {
            dest[n1++] = f->pushBack[--(f->pushed)];
        }
        if (n1 == 0) {
            return f->io->read(f, buf, len);
        } else {
            if (FILE_Eof(f)) {
                return n1;
            } else {
                int n2 = f->io->read(f, dest + n1, len - n1);
                if (n2 >= 0) {
                    return (n1 + n2);
                } else {
                    return n1;
                }
            }
        }
    } else {
        return (-1);
    }
}

/**
 * Reads data from stream. Returns True if all requested bytes have been
 * read from the stream, or False in case of an error, end of file condition
 * or partial read.
 */
Bool FILE_ReadAll(File * f, void * buf, int len)
{
    int n = FILE_Read(f, buf, len);
    if (n == len) {
        return True;
    } else {
        ASSERT(n < len);
        return False;
    }
}

/**
 * Pushes back some data that will be read from the file next time
 * FILE_Read is invoked. NOTE that the size of the pushback buffer
 * is very small. It's only guaranteed to be enough to push back
 * one character, so that one FILE_Ungetc call will always succeed.
 * Returns number of bytes actually pushed back to the file's buffer.
 * 
 * NOTE that the data are being placed to the pushback buffer starting
 * from the LAST byte in the provided buffer. That is, if this function
 * returns n, it means that the LAST n bytes have been pushed back.
 */
int FILE_PushBack(File * f, const void * buf, int len)
{
    int n = 0;
    if (CAN_READ(f)) {
        const I8u * src = (I8u*)buf;
        while (n < len && f->pushed < (int)sizeof(f->pushBack)) {
            f->pushBack[(f->pushed)++] = src[len - (n++) - 1];
        }
    }
    return n;
}

/**
 * Writes data to the stream. Returns the number of bytes actually written, 
 * which may be less than the buffer length if an error occurs. Returns (-1) 
 * if an error occurs and no data has been written.
 */
int FILE_Write(File * f, const void * buf, int len)
{
    ASSERT(f);
    if (CAN_WRITE(f)) {
        return f->io->write(f, buf, len);
    } else {
        return (-1);
    }
}

/**
 * Writes data to the stream. Returns True if the entire buffer has been 
 * written to the file, or False if less than len bytes have been written.
 */
Bool FILE_WriteAll(File * f, const void * buf, int len)
{
    ASSERT(f);
    if (CAN_WRITE(f)) {
        int n = f->io->write(f, buf, len);
        if (n == len) {
            return True;
        }
        ASSERT(n < len);
    }
    return False;
}

/**
 * Reads and discards the specified number of bytes from the stream.
 * Returns the number of bytes actually skipped 
 */
int FILE_Skip(File * f, int skip)
{
    int count = 0;
    if (skip > 0) {
        int n;
        int buf[64];

        /* read full buffers */
        while (skip > (int)sizeof(buf)) {
            n = FILE_Read(f, buf, sizeof(buf));
            if (n <= 0) {
                return count;
            } else {
                count += n;
                skip -= n;
            }
        }

        /* read the rest */
        n = FILE_Read(f, buf, skip);
        if (n > 0) count += n;
    }
    return count;
}

/**
 * Reads and discards the specified number of bytes from the stream.
 * Returns True if the requested number of bytes have been skipped.
 */
Bool FILE_SkipAll(File * f, int skip)
{
    int n = FILE_Skip(f, skip);
    if (n == skip) {
        return True;
    } else {
        ASSERT(n < skip);
        return False;
    }
}

/**
 * Prints formatted data to the stream. Returns the number of characters 
 * actually written to the stream, or (-1) if it fails.
 */
int FILE_Printf(File * f, Str format, ...)
{
    int nbytes;
    va_list va;
    va_start(va, format);
    nbytes = FILE_VaPrintf(f, format, va);
    va_end(va);        
    return nbytes;
}

/**
 * Prints formatted data to the stream. Returns the number of characters 
 * actually written to the stream, or (-1) if it fails.
 */
int FILE_VaPrintf(File * f, Str format, va_list va)
{
    int nbytes = -1;
    ASSERT(f);
    if (CAN_WRITE(f)) {
        if (STRBUF_FormatVa(&f->buf, format, va)) {
            nbytes = f->io->write(f, f->buf.s, f->buf.len);
        }
    }
    return nbytes;
}

/**
 * Writes a string to the stream. Returns True if string was written 
 * successfully, False if no data was written and MayBe if only part 
 * of the string was written to the stream.
 */
Bool FILE_Puts(File * f, Str s)
{
    ASSERT(f);
    ASSERT(s);
    if (CAN_WRITE(f)) {
        if (!*s) {
            return True;
        } else {
            size_t nbytes;
#ifdef UNICODE
            size_t n = wcstombs(NULL,s,0) + 1;
            char * bytes = NULL;
            if (((int)n) > 0) {

                /*
                 * use the string buffer for storing the result of 
                 * translating UNICODE string into multibyte 
                 */

                STRBUF_SetLength(&f->buf, 0);
                if (STRBUF_Alloc(&f->buf,(n+1)/2)) {
                    bytes = (char*)(f->buf.s);
                    n = wcstombs(bytes,s,n);
                    if (n == ((size_t)-1)) {
                        return False;
                    }
                    bytes[n] = 0;
                } else {
                    return False;
                }
            }
#else
            size_t n = StrLen(s);
            const char * bytes = s;
#endif /* ! UNICODE */

            nbytes = f->io->write(f, bytes, n);

#ifdef UNICODE
            if (n > 0) f->buf.s[0] = 0;
#endif /* UNICODE */
            if (nbytes > 0) { 
                if (nbytes == n) {
                    return True;
                } else {
                    return MayBe;   /* partial write */
                }
            }
        }
    }
    return False;
}

/**
 * Reads a string from the stream. False is returned to indicate an I/O
 * error, end-of-file condition or detection of binary data (i.e. a zero
 * has been received from the stream)
 */
Bool FILE_Gets(File * f, Char * buf, int len)
{
    Char * s = buf;
    if (!buf || !len || !CAN_READ(f)) return False;
    while (len-- > 1) {
        int c = FILE_Getc(f);
        if (c == 0) {
            FILE_Ungetc(f, (Char)c);
            break; /* binary data */
        }
        if ((c == EOF) || ((*s++ = (Char)c) == '\n')) break;
    }
    *s = 0;
    return BoolValue(s != buf || len == 0);
}

/**
 * Writes a character to the stream. Returns True on success, False on error
 */
Bool FILE_Putc(File * f, int c)
{
    Char s[2];
    s[0] = (Char)c;
    s[1] = 0;
    return FILE_Puts(f, s);
}

/**
 * Reads a character to the stream. Returns the character read,
 * or EOF in case of error.
 */
int FILE_Getc(File * f)
{
    ASSERT(f);
    if (CAN_READ(f)) {
#ifndef UNICODE
        char ch = 0;
        int nbytes = FILE_Read(f, &ch, 1);
        if (nbytes == 1) {
            return (((int)ch) & 0xff);
        }
#else  /* UNICODE */
        char mbs[5];
        int i, n = COUNT(mbs)-1;
        wchar_t ws[2];
        for (i=0; i<n; i++) {
            if (FILE_Read(f, mbs+i, 1) != 1) return EOF;
            mbs[i+1] = 0;
            if (mbstowcs(ws, mbs, COUNT(ws)) == 1) {
                return ws[0];
            }
        }
        ASSMSG("Invalid multibyte sequence?");
#endif /* UNICODE */
    }
    return EOF;
}

/**
 * Pushes c back to stream, where it is  available for subsequent read 
 * operations. Pushed-back characters will be returned in reverse order;
 * only one pushback is guaranteed.
 */
Bool FILE_Ungetc(File * f, Char c)
{
    Bool ok = False;
    ASSERT(f);
    if (CAN_READ(f)) {
#ifndef UNICODE
        ok = BoolValue(FILE_PushBack(f, &c, 1) == 1);
#else  /* UNICODE */
        int n;
        char mbs[COUNT(f->pushBack)+1];
        int avail = COUNT(f->pushBack) - f->pushed;
        wchar_t ws[2];
        ws[0] = c;
        ws[1] = 0;
        n = (int)wcstombs(mbs, ws, avail + 1);
        if (n > 0 && n <= avail) {
            ok = BoolValue(FILE_PushBack(f, mbs, n));
            ASSERT(ok);
        }
#endif /* UNICODE */
    }
    return ok;
}

/**
 * Flushes the stream. Returns True on success.
 */
Bool FILE_Flush(File * f)
{
    ASSERT(f);
    return ((f && (f->flags & FILE_IS_OPEN) && f->io->flush) ? 
            f->io->flush(f) : False);
}

/**
 * Tests if end of file has been reached
 */
Bool FILE_Eof(File * f)
{
    ASSERT(f);
    return ((f && (f->flags & FILE_IS_OPEN)) ? f->io->eof(f) : True);
}

/**
 * If this stream redirects (possibly modified) data to another file,
 * this method returns the target file. Examples of such streams are
 * ZIP and WRAP streams
 */
File * FILE_Target(File * f)
{
    ASSERT(f);
    if (f && f->io->target) {
        return f->io->target(f);
    }
    return NULL;
}

/**
 * Detach the File structure from the underlying socket/file descriptor, etc.
 * The File must have been created with one of the FILE_Attach* calls.
 */
void FILE_Detach(File * f)
{
    ASSERT(f);
    if (f) {
        if (f->flags & FILE_IS_ATTACHED) {
            f->io->detach(f);
            f->pushed = 0;
            f->flags &= (~(FILE_IS_OPEN | FILE_IS_ATTACHED));
        }
    }    
}

/**
 * Closes the file but doesn't deallocate the File structure. Any I/O
 * operation will return failure after this call.
 */
void FILE_Finish(File * f)
{
    ASSERT(f);
    if (f) {
        if (f->flags & FILE_IS_OPEN) {
            f->io->close(f);
            f->pushed = 0;
            f->flags &= (~FILE_IS_OPEN);
        }
    }    
}

/**
 * Closes the file. After this call returns, the File structure is 
 * no longer usable.
 */
void FILE_Close(File * f)
{
    ASSERT(f);
    if (f) {
        FileFree ff = f->io->free;
        FILE_Finish(f);
        STRBUF_Destroy(&f->name);
        STRBUF_Destroy(&f->buf);
        (*ff)(f);
    }
}

/*==========================================================================*
 *              U T I L I T I E S
 *==========================================================================*/

#define STACK_BUF_SIZE 1024  /* size of the buffer we allocate on stack */

/**
 * Copies the data from one stream and writes it into another. Returns number
 * of bytes copied, -1 if nothing has been written due to an error. There's 
 * no return value to detect partial copy.
 */
int FILE_Copy(File * in, File * out)
{
    int n, total = 0;
    I8u data[STACK_BUF_SIZE];
    while ((n = FILE_Read(in, data, sizeof(data))) > 0) {
        int nread = n;
        n = FILE_Write(out, data, nread);
        if (n > 0) {
            total += n;
            if (n == nread) {
                continue;
            }
        }
        break;
    }
    return ((total > 0) ? total : ((n == 0) ? 0 : -1));
}

/**
 * Copies no more than max bytes from one stream to another. Negative
 * max size means to copy all data until end of file. Returns number
 * of bytes successfully read from the input file and written to the
 * output file. Note that number of bytes read from the input file
 * may be greater than the number of bytes written to the output file.
 */
int FILE_CopyN(File * in, File * out, int max)
{
    if (max < 0) {
        return FILE_Copy(in, out);
    } else if (max == 0) {
        return FILE_Read(in, NULL, 0);
    } else {
        int n, total = 0;
        I8u data[STACK_BUF_SIZE];
        int chunk = MIN(STACK_BUF_SIZE,max);
        while ((n = FILE_Read(in, data, chunk)) > 0) {
            int nread = n;
            n = FILE_Write(out, data, nread);
            if (n > 0) {
                total += n;
                if (n == nread && max > total) {
                    chunk = MIN(STACK_BUF_SIZE,max-total);
                    continue;
                }
            }
            break;
        }
        return total;
    }
}

/**
 * Reads a line from the stream. Returns the pointer to the string buffer,
 * NULL if nothing was read from the file (that includes both end-of-line
 * and error conditions)
 */
Str FILE_ReadLine(File * in, StrBuf * sb) 
{
    STRBUF_Clear(sb);
    if (sb->alloc > 1 || STRBUF_Alloc(sb, 63)) {
        while (FILE_Gets(in, sb->s + sb->len, sb->alloc - sb->len)) {
            sb->len += StrLen(sb->s + sb->len);
            if (sb->len == 0) {
                ASSERT(False);
                return STRBUF_Text(sb);
            } else if (sb->s[sb->len-1] == '\n') {
                sb->s[--(sb->len)] = 0;
                if (sb->len > 0 && sb->s[sb->len-1] == '\r') {
                    sb->s[--(sb->len)] = 0;
                }
                return STRBUF_Text(sb);
            }

            /* we don't see '\n' in the end, continue reading */
            if (!STRBUF_Alloc(sb, sb->len * 2)) {
                return NULL;
            }
        }
        return ((sb->len > 0) ? STRBUF_Text(sb) : NULL);
    }
    return NULL;
}

/**
 * Reads the contents of the file into a buffer. Returns number of bytes
 * stored in the buffer. If max parameter is >= 0, reads no more than
 * max bytes
 */
int FILE_ReadData(File * in, Buffer * out, int max)
{
    I8u data[STACK_BUF_SIZE];
    int initial = BUFFER_Size(out);
    int bufsiz = initial;

    /* size of the first chunk to read */
    int chunk = ((max<0) ? STACK_BUF_SIZE : MIN(STACK_BUF_SIZE,max));
    while (BUFFER_EnsureCapacity(out, bufsiz + chunk, True)) {
        int maxsiz = MIN(out->alloc - bufsiz, chunk);
        int nbytes = FILE_Read(in, data, maxsiz);
        if (nbytes > 0) {
            /* this must succeed since memory has been pre-allocated */
            BUFFER_Put(out, data, nbytes, True);
            bufsiz += nbytes;
            ASSERT(bufsiz == BUFFER_Size(out));

            /* calculate the size of the next chunk */
            if (max < 0) {
                chunk = STACK_BUF_SIZE;
            } else {
                ASSERT(max >= nbytes);
                max -= nbytes;
                if (max == 0) {
                    break;
                } else {
                    chunk = MIN(STACK_BUF_SIZE,max);
                }
            }
        } else {
            break;
        }
    }
    return bufsiz - initial;
}

/**
 * Tests if this stream performs file based I/O. If this function return 
 * True, then the string returned by FILE_Name probably represents a file
 * on disk. Otherwise, it can be anything.
 */
Bool FILE_IsFileIO(File * f)
{
    ASSERT(f);
    while (f) {
        if (f->io->fileIO) {
            return True;
        }
        f = FILE_Target(f);
    }
    return False;
}

/**
 * The following functions write a 16/32/64-bit number to the stream
 * performing conversion it to the specified byte order if necessary.
 */
Bool FILE_WriteI16(File * f, I16u data, int endianess)
{
    data = DATA_Conv16(data, BYTE_ORDER, endianess);
    return FILE_WriteAll(f, &data, sizeof(data));
}

Bool FILE_WriteI32(File * f, I32u data, int endianess)
{
    data = DATA_Conv32(data, BYTE_ORDER, endianess);
    return FILE_WriteAll(f, &data, sizeof(data));
}

Bool FILE_WriteI64(File * f, I64u data, int endianess)
{
    data = DATA_Conv64(data, BYTE_ORDER, endianess);
    return FILE_WriteAll(f, &data, sizeof(data));
}

/**
 * The following functions read a 16/32/64-bit number to the stream
 * performing conversion from the specified byte order to the host
 * byte order if necessary.
 */
Bool FILE_ReadI16(File * f, I16u * data, int endianess)
{
    I16u tmp;
    ASSERT(data);
    if (FILE_ReadAll(f, &tmp, sizeof(*data))) {
        *data = DATA_Conv16(tmp, endianess, BYTE_ORDER);
        return True;
    }
    return False;
}

Bool FILE_ReadI32(File * f, I32u * data, int endianess)
{
    I32u tmp;
    ASSERT(data);
    if (FILE_ReadAll(f, &tmp, sizeof(*data))) {
        *data = DATA_Conv32(tmp, endianess, BYTE_ORDER);
        return True;
    }
    return False;
}

Bool FILE_ReadI64(File * f, I64u * data, int endianess)
{
    I64u tmp;
    ASSERT(data);
    if (FILE_ReadAll(f, &tmp, sizeof(*data))) {
        *data = DATA_Conv64(tmp, endianess, BYTE_ORDER);
        return True;
    }
    return False;
}

/**
 * The following functions write a 16/32/64-bit number to the stream
 * converting it to the network (big endian, MSB first) byte order
 * if necessary. It's functionally equivalent to FILE_WriteIxx(f,
 * data, BIG_ENDIAN) but these functions are a bit faster on big
 * endian systems because the byte order decision is made at compile
 * time.
 */
Bool FILE_WriteI16B(File * f, I16u data)
{
#if BYTE_ORDER != BIG_ENDIAN
    data = DATA_Conv16(data, BYTE_ORDER, BIG_ENDIAN);
#endif /* BYTE_ORDER != BIG_ENDIAN */
    return FILE_WriteAll(f, &data, sizeof(data));
}

Bool FILE_WriteI32B(File * f, I32u data)
{
#if BYTE_ORDER != BIG_ENDIAN
    data = DATA_Conv32(data, BYTE_ORDER, BIG_ENDIAN);
#endif /* BYTE_ORDER != BIG_ENDIAN */
    return FILE_WriteAll(f, &data, sizeof(data));
}

Bool FILE_WriteI64B(File * f, I64u data)
{
#if BYTE_ORDER != BIG_ENDIAN
    data = DATA_Conv64(data, BYTE_ORDER, BIG_ENDIAN);
#endif /* BYTE_ORDER != BIG_ENDIAN */
    return FILE_WriteAll(f, &data, sizeof(data));
}

/* The same, only for reading 16/32/64-bit numbers from big endian stream */
Bool FILE_ReadI16B(File * f, I16u * data)
{
#if BYTE_ORDER != BIG_ENDIAN
    I16u tmp;
    if (FILE_ReadAll(f, &tmp, sizeof(*data))) {
        *data = DATA_Conv16(tmp, BIG_ENDIAN, BYTE_ORDER);
        return True;
    }
    return False;
#else  /* BYTE_ORDER == BIG_ENDIAN */
    return FILE_ReadAll(f, data, sizeof(*data));
#endif /* BYTE_ORDER == BIG_ENDIAN */
}

Bool FILE_ReadI32B(File * f, I32u * data)
{
#if BYTE_ORDER != BIG_ENDIAN
    I32u tmp;
    if (FILE_ReadAll(f, &tmp, sizeof(*data))) {
        *data = DATA_Conv32(tmp, BIG_ENDIAN, BYTE_ORDER);
        return True;
    }
    return False;
#else  /* BYTE_ORDER == BIG_ENDIAN */
    return FILE_ReadAll(f, data, sizeof(*data));
#endif /* BYTE_ORDER == BIG_ENDIAN */
}

Bool FILE_ReadI64B(File * f, I64u * data)
{
#if BYTE_ORDER != BIG_ENDIAN
    I64u tmp;
    if (FILE_ReadAll(f, &tmp, sizeof(*data))) {
        *data = DATA_Conv64(tmp, BIG_ENDIAN, BYTE_ORDER);
        return True;
    }
    return False;
#else  /* BYTE_ORDER == BIG_ENDIAN */
    return FILE_ReadAll(f, data, sizeof(*data));
#endif /* BYTE_ORDER == BIG_ENDIAN */
}

/**
 * The following functions write a 16/32/64-bit number to the stream
 * converting it to little endian (LSB first) byte order if necessary.
 * It's functionally equivalent to FILE_WriteIxx(f, data, LITTLE_ENDIAN)
 * but it's a bit faster on little endian systems (such as Intel) because
 * the byte order decision is made at compile time.
 */
Bool FILE_WriteI16L(File * f, I16u data)
{
#if BYTE_ORDER != LITTLE_ENDIAN
    data = DATA_Conv16(data, BYTE_ORDER, LITTLE_ENDIAN);
#endif /* BYTE_ORDER == BIG_ENDIAN */
    return FILE_WriteAll(f, &data, 2);
}

Bool FILE_WriteI32L(File * f, I32u data)
{
#if BYTE_ORDER != LITTLE_ENDIAN
    data = DATA_Conv32(data, BYTE_ORDER, LITTLE_ENDIAN);
#endif /* BYTE_ORDER == BIG_ENDIAN */
    return FILE_WriteAll(f, &data, 4);
}

Bool FILE_WriteI64L(File * f, I64u data)
{
#if BYTE_ORDER != LITTLE_ENDIAN
    data = DATA_Conv64(data, BYTE_ORDER, LITTLE_ENDIAN);
#endif /* BYTE_ORDER == BIG_ENDIAN */
    return FILE_WriteAll(f, &data, 2);
}

/* The same, only for reading numbers from little endian stream */
Bool FILE_ReadI16L(File * f, I16u * data)
{
#if BYTE_ORDER != LITTLE_ENDIAN
    I16u tmp;
    if (FILE_ReadAll(f, &tmp, sizeof(*data))) {
        *data = DATA_Conv16(tmp, LITTLE_ENDIAN, BYTE_ORDER);
        return True;
    }
    return False;
#else  /* BYTE_ORDER == LITTLE_ENDIAN */
    return FILE_ReadAll(f, data, sizeof(*data));
#endif /* BYTE_ORDER == LITTLE_ENDIAN */
}

Bool FILE_ReadI32L(File * f, I32u * data)
{
#if BYTE_ORDER != LITTLE_ENDIAN
    I32u tmp;
    if (FILE_ReadAll(f, &tmp, sizeof(*data))) {
        *data = DATA_Conv32(tmp, LITTLE_ENDIAN, BYTE_ORDER);
        return True;
    }
    return False;
#else  /* BYTE_ORDER == LITTLE_ENDIAN */
    return FILE_ReadAll(f, data, sizeof(*data));
#endif /* BYTE_ORDER == LITTLE_ENDIAN */
}

Bool FILE_ReadI64L(File * f, I64u * data)
{
#if BYTE_ORDER != LITTLE_ENDIAN
    I64u tmp;
    if (FILE_ReadAll(f, &tmp, sizeof(*data))) {
        *data = DATA_Conv64(tmp, LITTLE_ENDIAN, BYTE_ORDER);
        return True;
    }
    return False;
#else  /* BYTE_ORDER == LITTLE_ENDIAN */
    return FILE_ReadAll(f, data, sizeof(*data));
#endif /* BYTE_ORDER == LITTLE_ENDIAN */
}

#ifndef __KERNEL__
/* Assumptions we make about the size of floating point data types */
COMPILE_ASSERT(sizeof(float) == sizeof(I32u));
COMPILE_ASSERT(sizeof(double) == sizeof(I64u));

/**
 * The following functions write a 32/64-bit (single/double precision)
 * floating-point number to the stream converting it to the specified
 * byte order if necessary.
 */
Bool FILE_WriteF32(File * f, float data, int endianess)
{
    I32u tmp = DATA_Conv32(*((I32u*)&data), BYTE_ORDER, endianess);
    return FILE_WriteAll(f, &tmp, 4);
}

Bool FILE_WriteF64(File * f, double data, int endianess)
{
    I64u tmp = DATA_Conv64(*((I64u*)&data), BYTE_ORDER, endianess);
    return FILE_WriteAll(f, &tmp, 8);
}

/* Reading numbers from a stream */
Bool FILE_ReadF32(File * f, float * data, int endianess)
{
    I32u tmp;
    ASSERT(data);
    if (FILE_ReadAll(f, &tmp, sizeof(tmp))) {
        *((I32u*)data) = DATA_Conv32(tmp, endianess, BYTE_ORDER);
        return True;
    }
    return False;
}

Bool FILE_ReadF64(File * f, double * data, int endianess)
{
    I64u tmp;
    ASSERT(data);
    if (FILE_ReadAll(f, &tmp, sizeof(tmp))) {
        *((I64u*)data) = DATA_Conv64(tmp, endianess, BYTE_ORDER);
        return True;
    }
    return False;
}

/**
 * The following functions write a 32/64-bit (single/double precision)
 * floating-point number to the stream converting it to the network
 * (big endian, MSB first) byte order if necessary. It's functionally
 * equivalent to FILE_WriteFxx(f,data,BIG_ENDIAN) but it's a bit faster
 * on big endian systems because the byte order decision is made at compile
 * time.
 */
Bool FILE_WriteF32B(File * f, float data)
{
    I32u tmp = (*((I32u*)&data));
#if BYTE_ORDER != BIG_ENDIAN
    tmp = DATA_Conv32(tmp, BYTE_ORDER, BIG_ENDIAN);
#endif /* BYTE_ORDER == BIG_ENDIAN */
    return FILE_WriteAll(f, &tmp, 4);
}

Bool FILE_WriteF64B(File * f, double data)
{
    I64u tmp = (*((I64u*)&data));
#if BYTE_ORDER != BIG_ENDIAN
    tmp = DATA_Conv64(tmp, BYTE_ORDER, BIG_ENDIAN);
#endif /* BYTE_ORDER == BIG_ENDIAN */
    return FILE_WriteAll(f, &tmp, 8);
}

/* The same, only for reading numbers from big endian stream */
Bool FILE_ReadF32B(File * f, float * data)
{
#if BYTE_ORDER != BIG_ENDIAN
    I32u tmp;
    ASSERT(data);
    if (FILE_ReadAll(f, &tmp, sizeof(tmp))) {
        *((I32u*)data) = DATA_Conv32(tmp, BIG_ENDIAN, BYTE_ORDER);
        return True;
    }
    return False;
#else  /* BYTE_ORDER == BIG_ENDIAN */
    return FILE_ReadAll(f, data, sizeof(*data));
#endif /* BYTE_ORDER == BIG_ENDIAN */
}

Bool FILE_ReadF64B(File * f, double * data)
{
#if BYTE_ORDER != BIG_ENDIAN
    I64u tmp;
    ASSERT(data);
    if (FILE_ReadAll(f, &tmp, sizeof(tmp))) {
        *((I64u*)data) = DATA_Conv64(tmp, BIG_ENDIAN, BYTE_ORDER);
        return True;
    }
    return False;
#else  /* BYTE_ORDER == BIG_ENDIAN */
    return FILE_ReadAll(f, data, sizeof(*data));
#endif /* BYTE_ORDER == BIG_ENDIAN */
}

/**
 * The following functions write a 32/64-bit (single/double precision)
 * floating-point number to the stream converting it to little endian
 * (LMSB first) byte order if necessary. It's functionally equivalent
 * to FILE_WriteFxx(f,data,LITTLE_ENDIAN) but it's a bit faster on little
 * endian systems because the byte order decision is made at compile time.
 */
Bool FILE_WriteF32L(File * f, float data)
{
    I32u tmp = (*((I32u*)&data));
#if BYTE_ORDER != LITTLE_ENDIAN
    tmp = DATA_Conv32(tmp, BYTE_ORDER, LITTLE_ENDIAN);
#endif /* BYTE_ORDER == BIG_ENDIAN */
    return FILE_WriteAll(f, &tmp, 4);
}

Bool FILE_WriteF64L(File * f, double data)
{
    I64u tmp = (*((I64u*)&data));
#if BYTE_ORDER != LITTLE_ENDIAN
    tmp = DATA_Conv64(tmp, BYTE_ORDER, LITTLE_ENDIAN);
#endif /* BYTE_ORDER == BIG_ENDIAN */
    return FILE_WriteAll(f, &tmp, 8);
}

/* The same, only for reading numbers from little endian stream */
Bool FILE_ReadF32L(File * f, float * data)
{
#if BYTE_ORDER != LITTLE_ENDIAN
    I32u tmp;
    ASSERT(data);
    if (FILE_ReadAll(f, &tmp, sizeof(tmp))) {
        *((I32u*)data) = DATA_Conv32(tmp, LITTLE_ENDIAN, BYTE_ORDER);
        return True;
    }
    return False;
#else  /* BYTE_ORDER == LITTLE_ENDIAN */
    return FILE_ReadAll(f, data, sizeof(*data));
#endif /* BYTE_ORDER == LITTLE_ENDIAN */
}

Bool FILE_ReadF64L(File * f, double * data)
{
#if BYTE_ORDER != LITTLE_ENDIAN
    I64u tmp;
    ASSERT(data);
    if (FILE_ReadAll(f, &tmp, sizeof(tmp))) {
        *((I64u*)data) = DATA_Conv64(tmp, LITTLE_ENDIAN, BYTE_ORDER);
        return True;
    }
    return False;
#else  /* BYTE_ORDER == LITTLE_ENDIAN */
    return FILE_ReadAll(f, data, sizeof(*data));
#endif /* BYTE_ORDER == LITTLE_ENDIAN */
}

#endif /* __KERNEL__ */
